//// `embeds/files` generates Gleam modules containing constants and zero-argument
//// functions based on the contents of files.
////
//// This module is intended to be run as a CLI tool during your build process:
//// ```sh
//// gleam run -m embeds/files -- --help
//// ```
////
//// The functions exposed in this module can also be used directly to support
//// advanced use cases inside custom pre-build scripts.

import argv
import embeds/generate.{type Code, type Data, type Definition, type Error}
import filepath
import gleam/bool
import gleam/int
import gleam/io
import gleam/list
import gleam/option.{type Option, None, Some}
import gleam/regex
import gleam/result
import gleam/string
import glint.{type Command}
import simplifile.{type FileError}

pub type Options {
  Options(
    /// A list of sources to scan for files.
    /// See the `Source` documentation for more info.
    sources: List(Source),
    /// The header to place on the top of all modules
    header: String,
    /// A printer is a function that turns a file into some Gleam code.
    /// You can use the default `constants_printer` to produce String or BitArray literals.
    print: fn(String, Data) -> Code,
    /// If force is enabled, all files will be regenerated, regardless of whether
    /// or not their timestamp has changed.
    force: Bool,
  )
}

/// A Source, describing how to map a directory structure to Gleam module names.
pub type Source {
  Source(
    /// The source directory, where your asset files live.
    src: String,
    /// The base name of the generated module.
    /// The module name will be used as a base for all files in `src`, and nested
    /// directories will generate nested modules. (see `flatten` on how to tweak this)
    /// Multiple sources can re-use the same module name.
    module: String,
    /// how many levels of directories to ignore before starting to group things into nested modules,
    /// which is sometimes useful to "skip" some diretories first.
    flatten: Flatten,
    /// The maximum amount of levels to traverse before stopping looking for files.
    /// This setting is independent of `flatten` and will instead work based on the directory tree.
    /// A value of 0 means to only look at the given `src` directory.
    /// A None value represents no limit.
    max_depth: Option(Int),
    /// A filter function called on file and directory names.
    /// If the function returns `False`, the directory or file will be skipped.
    /// Set to `default_filter` to filter out hidden files and directories.
    filter: fn(Bool, String) -> Bool,
  )
}

pub type Flatten {
  /// Don't flatten. Every directory will generate another sub-module.
  DontFlatten
  /// Flatten all files, resulting in a single Gleam module
  FlattenAll
  /// Only flatten the first N directory levels
  FlattenFirst(Int)
}

/// A default filter function to use inside sources.
/// Filters out hidden files and directories.
pub fn default_filter(_is_dir: Bool, path: String) -> Bool {
  let is_hidden = filepath.base_name(path) |> string.starts_with(".")
  is_hidden == False
}

/// Generate Gleam modules, containing constants for all found files, given the sources.
/// Exported for easier integration with other build tools.
///
/// ## Example
///
/// ```gleam
/// import embeds/files
/// import embeds/generate
/// import gleam/option
/// import gleam/io
///
/// pub fn main() {
///     let result = files.generate(files.Options(
///         sources: [files.Source(
///             src: "./assets",
///             module: "assets",
///             flatten: files.FlattenAll,
///             max_depth: option.None,
///             filter: files.default_filter,
///         )],
///
///         print: generate.default_print,
///         header: generate.default_header,
///         force: False,
///     ))
///
///     case result {
///        Ok(Nil) -> Nil
///        Error(errs) -> io.println_error(generate.describe_errors(errs))
///     }
/// }
/// ```
pub fn generate(options: Options) -> Result(Nil, List(Error)) {
  use <- generate.guard_gleam_project()

  let Options(sources: sources, header: header, print: print, force: force) =
    options

  // collect files
  use defs <- result.try({
    use spec <- traverse_map(sources)
    get_files(spec)
    |> result.map_error(fn(e) { [generate.CouldNotListFiles(spec.src, e)] })
  })

  generate.generate(list.flatten(defs), header, print, force)
}

//
// -- COMMAND LINE INTERFACE
//

/// runs generate using command line arguments to fill out the arguments.
pub fn main() {
  glint.new()
  |> glint.with_name("embeds/files")
  |> glint.as_module()
  |> glint.pretty_help(glint.default_pretty_help())
  |> glint.add(at: [], do: generate_cmd())
  |> glint.run(argv.load().arguments)
}

fn generate_cmd() -> Command(Nil) {
  use <- glint.command_help(
    "Generate Gleam modules containing String/BitArray constants of static files.\n\n\n"
    <> "By default, looks for a directory called assets, and generates a Gleam module called assets. The directory structure is preserved in the module structure.\n\n\n"
    <> "ARGS can be  a list of any of the following:\n\n\n"
    <> "- ./path/to/files:\n\n"
    <> "    List all files in ./path/to/files, generating a Gleam module called `files` (following the directory name).\n\n"
    <> "- assets:./path/to/files\n\n"
    <> "    Rename the top-level module to `assets`\n\n",
  )

  use force <- glint.flag(
    glint.bool_flag("force")
    |> glint.flag_default(False)
    |> glint.flag_help("Do not skip generating based on timestamps"),
  )

  use flatten_bool <- glint.flag(
    glint.bool_flag("flatten")
    |> glint.flag_default(False)
    |> glint.flag_help(
      "Do not generate nested modules reflecting the directory tree",
    ),
  )

  use flatten_int <- glint.flag(
    glint.int_flag("strip")
    |> glint.flag_default(0)
    |> glint.flag_help(
      "Like flatten, but only strip the first N levels from the directory tree. If both are given, flatten takes precedence.",
    ),
  )

  use max_depth <- glint.flag(
    glint.int_flag("max-depth")
    |> glint.flag_help(
      "The maximum level of nesting before stopping looking for files.",
    ),
  )

  use filter <- glint.flag(
    glint.string_flag("filter")
    |> glint.flag_help(
      "A regular expression that needs to match the relative file path.",
    ),
  )

  use _, args, flags <- glint.command()

  let assert Ok(force) = force(flags)
  let assert Ok(flatten_bool) = flatten_bool(flags)
  let assert Ok(flatten_int) = flatten_int(flags)
  let max_depth = max_depth(flags) |> option.from_result
  let filter = filter(flags) |> option.from_result

  let flatten = case flatten_bool, flatten_int {
    True, _ -> FlattenAll
    False, 0 -> DontFlatten
    False, n -> FlattenFirst(n)
  }

  let result = {
    use filter <- result.try(case filter {
      Some(filter) ->
        case regex.from_string(filter) {
          Ok(regex) ->
            Ok(fn(is_dir, path) {
              case default_filter(is_dir, path), is_dir {
                False, _ -> False
                True, False -> regex.check(regex, path)
                True, True -> True
              }
            })

          Error(regex.CompileError(error: error, ..)) ->
            Error([generate.InvalidArgument("filter", filter, error)])
        }
      None -> Ok(default_filter)
    })

    // parse path specs from args
    use path_specs <- result.try(case args {
      [] ->
        Ok([
          Source(
            src: "./assets",
            module: "assets",
            flatten: flatten,
            max_depth: max_depth,
            filter: filter,
          ),
        ])
      _ -> {
        use spec <- traverse_map(args)
        parse_path_spec(spec, flatten, max_depth, filter)
        |> result.replace_error([
          generate.InvalidArgument("", spec, "Invalid spec string"),
        ])
      }
    })

    let options =
      Options(
        sources: path_specs,
        header: generate.default_header,
        force: force,
        print: generate.default_print,
      )

    generate(options)
  }

  case result {
    Ok(Nil) -> Nil
    Error(errs) -> io.println_error(generate.describe_errors(errs))
  }

  Nil
}

/// Parse a `Source` given as a command-line argument into our Gleam structure.
fn parse_path_spec(
  spec: String,
  flatten: Flatten,
  max_depth: Option(Int),
  filter: fn(Bool, String) -> Bool,
) -> Result(Source, Nil) {
  case string.split(spec, ":") {
    [src] -> {
      use src <- result.map(filepath.expand(src))
      Source(
        src: src,
        module: generate.to_gleam_module_name(src),
        flatten: flatten,
        max_depth: max_depth,
        filter: filter,
      )
    }
    [module, src] -> {
      use src <- result.map(filepath.expand(src))
      Source(
        src: src,
        module: generate.to_gleam_module_name(module),
        flatten: flatten,
        max_depth: max_depth,
        filter: filter,
      )
    }
    _ -> Error(Nil)
  }
}

//
// -- ADVANCED
//

/// Find all files inside a `Source`.
/// You usually do not need to call this function yourself.
pub fn get_files(src: Source) -> Result(List(Definition), FileError) {
  do_get_files(src, "", 0)
}

fn do_get_files(
  src: Source,
  nested: String,
  depth: Int,
) -> Result(List(Definition), FileError) {
  use contents <- result.try(
    simplifile.read_directory(filepath.join(src.src, nested)),
  )

  use acc, content <- list.try_fold(over: contents, from: [])
  let relative_path = filepath.join(nested, content)
  let absolute_path = filepath.join(src.src, relative_path)

  use info <- result.try(simplifile.file_info(absolute_path))
  // TODO: erlang exposes a "type" field, why doesn't simplifile?
  let is_file = int.bitwise_and(info.mode, 0o100000) > 0
  let is_dir = int.bitwise_and(info.mode, 0o40000) > 0

  use <- bool.guard(!src.filter(is_dir, relative_path), Ok(acc))

  case is_file, is_dir, src.max_depth {
    True, _, _ -> {
      let #(module_name, var_name) = get_file_name(src, relative_path)

      let definition =
        generate.Definition(
          name: absolute_path,
          last_modified_seconds: info.mtime_seconds,
          module: module_name,
          var_name: var_name,
          read: fn() { read_file(absolute_path) },
        )

      Ok([definition, ..acc])
    }
    _, True, None -> {
      use nested_files <- result.try(do_get_files(src, relative_path, depth + 1))
      Ok(list.append(acc, nested_files))
    }

    _, True, Some(max_depth) if depth < max_depth -> {
      use nested_files <- result.try(do_get_files(src, relative_path, depth + 1))
      Ok(list.append(acc, nested_files))
    }

    _, _, _ -> Ok(acc)
  }
}

//
// -- INTERNAL
//

fn read_file(path: String) -> Result(generate.Data, String) {
  case simplifile.read_bits(path) {
    Ok(bits) -> Ok(generate.bits_to_data(bits))
    Error(err) -> Error(simplifile.describe_error(err))
  }
}

/// Get the module and variable name for a file, based on its relative path inside a source.
fn get_file_name(src: Source, relative_path: String) -> #(String, String) {
  let assert Ok(#(module_parts, file_name)) =
    relative_path
    |> string.split("/")
    |> list.filter(fn(part) { part != "" && part != "." && part != ".." })
    |> split_last

  let var_name =
    file_name
    |> filepath.strip_extension
    |> generate.to_gleam_identifier

  let module_name =
    generate.to_gleam_module_name(case src.flatten {
      FlattenAll -> src.module
      FlattenFirst(n) ->
        module_parts
        |> list.drop(n)
        |> list.prepend(src.module)
        |> string.join("/")

      DontFlatten ->
        module_parts
        |> list.prepend(src.module)
        |> string.join("/")
    })

  #(module_name, var_name)
}

//
// -- HELPERS
//

fn split_last(list: List(a)) -> Result(#(List(a), a), Nil) {
  case list {
    [] -> Error(Nil)
    [first, ..list] -> Ok(do_split_last(list, [], first))
  }
}

fn do_split_last(list: List(a), init: List(a), last: a) -> #(List(a), a) {
  case list {
    [] -> #(list.reverse(init), last)
    [first, ..rest] -> do_split_last(rest, [last, ..init], first)
  }
}

fn traverse_map(
  list: List(a),
  body: fn(a) -> Result(ok, List(err)),
) -> Result(List(ok), List(err)) {
  do_traverse_map(list, body, [], [])
}

fn do_traverse_map(
  list: List(a),
  body: fn(a) -> Result(ok, List(err)),
  oks: List(ok),
  errs: List(err),
) -> Result(List(ok), List(err)) {
  case list {
    [] ->
      case errs {
        [] -> Ok(list.reverse(oks))
        _ -> Error(list.reverse(errs))
      }
    [head, ..tail] ->
      case body(head) {
        Ok(ok) -> do_traverse_map(tail, body, [ok, ..oks], errs)
        Error(err) -> do_traverse_map(tail, body, oks, list.append(err, errs))
      }
  }
}
