//// This module exposes the internal/shared code for both `embeds/files` as
//// well as `embeds/env`.
////
//// It exposes some common type aliases and functions, and provides a
//// lower-level API to generate module files from a set of `Definitions`.
////
//// I recommend starting with the `embeds/files` and/or `embeds/env` modules
//// first before diving into this one!

import filepath
import gleam/bit_array
import gleam/bool
import gleam/list
import gleam/regex
import gleam/result
import gleam/string
import simplifile.{type FileError}

/// A simple module comment which is used as the header by the CLI
pub const default_header = "//// Generated by `embeds`, manual edits are futile"

/// convert data to code, by simply calling string.inspect/bit_array.inspect.
pub fn default_print(_name: String, data: Data) -> Code {
  let code = case data {
    Text(text) -> string.inspect(text)
    Binary(bits) -> bit_array.inspect(bits)
  }

  Constant(code)
}

//
// -- ERROR
//

/// the internal Error type returned by all embeds functions.
pub type Error {
  /// the current working directory is not a Gleam project, and has no `gleam.toml` file.
  NotInsideGleamProject
  /// a command-line argument could not be parsed or is invalid.
  InvalidArgument(name: String, value: String, reason: String)
  /// A simplifile error occurred while listing all files.
  CouldNotListFiles(path: String, FileError)
  /// A simplifile error occurred while trying to read an asset file.
  CouldNotRead(name: String, message: String)
  /// Multiple files which would produce the same constant in the same moduele have been found.
  DuplicateVarName(String, String)
  /// The printer returned an error that should be reported.
  PrinterError(name: String, message: String)
  /// A simplifile error while writing to the output module.
  WriteError(path: String, FileError)
}

/// Turn an `Error` into a human-readable string.
pub fn describe_error(err: Error) -> String {
  case err {
    NotInsideGleamProject ->
      "generate needs to be called inside the root directory of your Gleam project!"
    InvalidArgument(name, value, reason) ->
      "Invalid argument " <> name <> " (" <> value <> "): " <> reason
    CouldNotListFiles(path, e) ->
      "Could not list files from "
      <> path
      <> ": "
      <> simplifile.describe_error(e)
    CouldNotRead(path, e) -> "Could not read " <> path <> ": " <> e
    DuplicateVarName(a, b) ->
      "I have found 2 source definitions "
      <> a
      <> " and "
      <> b
      <> " - both would produce the same constant in the same module!"
    PrinterError(path, e) ->
      "The printer for " <> path <> " returned an error: " <> e

    WriteError(path, e) ->
      "Error writing module " <> path <> ": " <> simplifile.describe_error(e)
  }
}

/// turn multiple erros into a human-readable string.
pub fn describe_errors(errs: List(Error)) -> String {
  string.join(list.map(errs, describe_error), "\n")
}

/// enforce that the current working directory is a gleam project
/// (by checking a gleam.toml exists).
/// Automatically called by `generate`, but you may wish to exit earlier.
/// Designed to be used with `use`.
pub fn guard_gleam_project(
  f: fn() -> Result(ok, List(Error)),
) -> Result(ok, List(Error)) {
  case simplifile.is_file("gleam.toml") {
    Ok(True) -> f()
    _ -> Error([NotInsideGleamProject])
  }
}

//
// -- DATA
//

/// Data represents the contents of some file or other external source,
/// and can either be UTF-8 `Text`, or `Binary`.
pub type Data {
  Text(String)
  Binary(BitArray)
}

/// Try to convert a BitArray to `Text` first , returning `Binary` if the
/// conversion fails.
pub fn bits_to_data(bits: BitArray) -> Data {
  case bit_array.to_string(bits) {
    Ok(text) -> Text(text)
    Error(_) -> Binary(bits)
  }
}

/// Convert some data back to a BitArray.
pub fn data_to_bits(data: Data) -> BitArray {
  case data {
    Text(text) -> bit_array.from_string(text)
    Binary(bits) -> bits
  }
}

//
// -- CODE
//

/// Code represents the rendered string of some Gleam expression.
///
/// Because some expressions can't be constants, you can also choose to also
/// print function bodies instead.
pub type Code {
  /// Skip generating code for this symbol.
  Skip
  /// Codegen produced some error that should be reported to the user.
  GenerateError(String)
  /// Produce a constant definition in Gleam, like
  /// ```gleam
  /// pub const var_name = ...
  /// ```
  Constant(String)
  /// Produce a zero-argument function definition in Gleam, like
  /// ```gleam
  /// pub fn var_name() {
  ///     ...
  /// }
  /// You can use this constructor instead of `Constant` when your code contains
  /// anonymous functions, function calls, or other expressions that are not
  /// allowed inside of constants.
  FunctionBody(String)
}

/// Helper function to convert a Result into either a `Constant` or `GenerateError`.
pub fn result_to_const(result: Result(String, String)) -> Code {
  case result {
    Ok(code) -> Constant(code)
    Error(err) -> GenerateError(err)
  }
}

/// Helper function to convert a Result into either a `Constant` or skip generating on error.
pub fn skip_or_const(result: Result(String, Nil)) -> Code {
  case result {
    Ok(code) -> Constant(code)
    Error(Nil) -> Skip
  }
}

/// Helper function to convert a Result into either a `FunctionBody` or `GenerateError`.
pub fn result_to_fn(result: Result(String, String)) -> Code {
  case result {
    Ok(code) -> FunctionBody(code)
    Error(err) -> GenerateError(err)
  }
}

/// Helper function to convert a Result into either a `FunctionBody` or skip generating on error.
pub fn skip_or_fn(result: Result(String, Nil)) -> Code {
  case result {
    Ok(code) -> FunctionBody(code)
    Error(Nil) -> Skip
  }
}

//
// -- DEFINITION
//

/// A single codegen unit, potentially producing a single `Code` statement
/// (so either a constant definition or a function.)
pub type Definition {
  Definition(
    /// A human-identifyable `name` for this definition used in error messages.
    /// Usually the file name or name of the environment variable.
    name: String,
    /// Files might get big, and definitions in general might be fetched from anywhere,
    /// so reading is deferred using this closure.
    read: fn() -> Result(Data, String),
    /// A unix timestamp indicating the last modification time in seconds since
    /// Jan 01, 1970 for the source file of this definition.
    ///
    /// If a module only contains definitions that were modified before the
    /// module was last generated, generation will be skipped.
    ///
    /// You can savely just set this value to 0 if you don't have a good value
    /// available.
    last_modified_seconds: Int,
    /// The name of the module this definition belongs to.
    /// Note that no sanitization takes place, and `module` should be a valid
    /// Gleam module name.
    module: String,
    /// The name of the target constant or function name.
    /// Note that no sanitization takes place, and `var_name` should be a valid
    /// Gleam identifier.
    var_name: String,
  )
}

//
// -- ESCAPING
//

/// Ensures the given string is a valid Gleam identifier by replacing invalid
/// characters by underscores.
///
/// Empty strings return "empty", numbers will be prefixed with "n".
///
/// If the resulting variable name would be a reserved keyword, an additional
/// underscore will be added at the end.
pub fn to_gleam_identifier(str: String) -> String {
  let assert Ok(re) = regex.from_string("[^a-zA-Z0-9]+")
  let var_name =
    str
    // first replace with " " such that we can use string.trim
    |> regex.replace(re, _, " ")
    // identifiers cannot start with an underscore, and should not end with one
    |> string.trim
    |> string.replace(each: " ", with: "_")
    |> string.lowercase

  // special cases:
  // - identifiers cannot start with a number
  // - identifiers cannot be empty
  // - identifiers cannot be a reserved keyword
  case var_name {
    // numbers
    "0" <> _
    | "1" <> _
    | "2" <> _
    | "3" <> _
    | "4" <> _
    | "5" <> _
    | "6" <> _
    | "7" <> _
    | "8" <> _
    | "9" <> _ -> "n" <> var_name

    // empty
    "" -> "empty"

    // keywords
    // https://github.com/gleam-lang/gleam/blob/main/compiler-core/src/parse/lexer.rs#L21
    "as"
    | "assert"
    | "auto"
    | "case"
    | "const"
    | "delegate"
    | "derive"
    | "echo"
    | "else"
    | "fn"
    | "if"
    | "implement"
    | "import"
    | "let"
    | "macro"
    | "opaque"
    | "panic"
    | "pub"
    | "test"
    | "todo"
    | "type"
    | "use" -> var_name <> "_"

    _ -> var_name
  }
}

/// Ensures a path is a valid Gleam module path, by making sure that every piece
/// is a valid Gleam identifier.
///
/// Unix `/` and Windows `\\` are both treated as module separators.
///
/// **NOTE**:The top-level module is not allowed to be called `gleam`, but this
///  function does not enforce this!
pub fn to_gleam_module_name(path: String) -> String {
  let module =
    string.join(
      {
        use part <- list.filter_map(
          path
          |> string.replace(each: "\\", with: "/")
          |> string.split(on: "/"),
        )

        case part {
          "." | ".." | "" -> Error(Nil)
          _ -> Ok(to_gleam_identifier(part))
        }
      },
      "/",
    )

  case module {
    "" -> "assets"
    _ -> module
  }
}

/// Turn a Gleam module name into the path to its source file.
fn module_file(module: String) -> String {
  filepath.join("src", module <> ".gleam")
}

//
// -- GENERATE
//

/// Generates module files based on a list of Definitions.
pub fn generate(
  defs: List(Definition),
  header: String,
  print: fn(String, Data) -> Code,
  force: Bool,
) -> Result(Nil, List(Error)) {
  generate_modules(defs, header, print, force, get_mtime, write)
}

/// DI version used in tests to avoid interacting with the file system for real.
@internal
pub fn generate_modules(
  defs: List(Definition),
  header: String,
  print: fn(String, Data) -> Code,
  force: Bool,
  get_mtime: fn(String) -> Result(Int, FileError),
  write: fn(String, String) -> Result(Nil, FileError),
) {
  // group by output module, make sure all variables are unique.
  use _x <- result.map({
    use module_specs <- result.try(
      defs
      |> gather_by(fn(file) { file.module })
      |> traverse_map(fn(pair) {
        case enforce_unique_by(pair.1, fn(file) { file.var_name }) {
          Ok(files) -> {
            let files =
              files
              |> list.sort(fn(a, b) { string.compare(a.var_name, b.var_name) })

            Ok(#(pair.0, files))
          }

          Error(conflict) ->
            Error([DuplicateVarName({ conflict.0 }.name, { conflict.1 }.name)])
        }
      }),
    )

    // write modules
    use #(module_name, defs) <- traverse_map(module_specs)
    let module_file = module_file(module_name)

    // skip module if no files have changed.
    // TODO: what about deleted files?
    let needs_update = case get_mtime(module_file) {
      Ok(module_ts) -> {
        use Definition(last_modified_seconds: file_ts, ..) <- list.any(defs)
        file_ts <= 0 || file_ts >= module_ts
      }

      Error(_) -> True
    }

    // NOTE: returning Error([]) is equivalent to a "continue" in procedual loops!
    use <- bool.guard(!force && !needs_update, Error([]))

    use contents <- result.try(generate_module_file(defs, header, print))

    use _ <- try(
      write(module_file, contents)
      |> result.map_error(fn(e) { WriteError(module_file, e) }),
    )

    Ok(Nil)
  })

  Nil
}

/// Generate a the source text for a single module file.
/// the `module` field of the definitions will be ignored.
@internal
pub fn generate_module_file(
  definitions: List(Definition),
  header: String,
  print: fn(String, Data) -> Code,
) -> Result(String, List(Error)) {
  use definitions <- result.try({
    use definition <- traverse_map(definitions)

    use data <- try(
      definition.read()
      |> result.map_error(fn(e) { CouldNotRead(definition.name, e) }),
    )

    let code = print(definition.name, data)

    case code {
      Skip -> Error([])
      GenerateError(err) -> Error([PrinterError(definition.name, err)])
      Constant(code) ->
        Ok(
          "/// "
          <> definition.name
          <> "\n"
          <> "pub const "
          <> definition.var_name
          <> " = "
          <> code,
        )

      FunctionBody(code) ->
        Ok(
          "/// "
          <> definition.name
          <> "\n"
          <> "pub fn "
          <> definition.var_name
          <> "() {\n"
          <> code
          <> "\n}",
        )
    }
  })

  case definitions, header {
    [], _ -> Ok("")
    _, "" -> Ok(string.join(definitions, "\n\n"))
    _, _ -> Ok(header <> "\n\n\n" <> string.join(definitions, "\n\n"))
  }
}

/// Actual implementation for `generate_modules`
fn get_mtime(path: String) {
  use info <- result.map(simplifile.file_info(path))
  info.mtime_seconds
}

/// Actual implementation for `generate_modules`
fn write(path: String, contents: String) -> Result(Nil, FileError) {
  use _ <- result.try(
    path
    |> filepath.directory_name
    |> simplifile.create_directory_all,
  )

  simplifile.write(to: path, contents: contents)
}

//
// -- HELPERS
//

/// Like dict.group >> dict.to_list, but stable wrt. order.
fn gather_by(
  list: List(elem),
  by to_key: fn(elem) -> key,
) -> List(#(key, List(elem))) {
  do_gather_by(list, to_key, [])
}

fn do_gather_by(
  list: List(elem),
  to_key: fn(elem) -> key,
  acc: List(#(key, List(elem))),
) -> List(#(key, List(elem))) {
  case list {
    [] -> list.reverse(acc)
    [head, ..tail] -> {
      let key = to_key(head)
      let #(matching, remaining) =
        list.partition(tail, fn(other) { to_key(other) == key })
      do_gather_by(remaining, to_key, [#(key, [head, ..matching]), ..acc])
    }
  }
}

/// Make sure all items in a list are unique using a key-generating function, or
/// return the first tuple of elements that have matching keys if they are not.
fn enforce_unique_by(
  list: List(a),
  by to_key: fn(a) -> b,
) -> Result(List(a), #(a, a)) {
  use acc, elem <- list.try_fold(over: list, from: [])
  let key = to_key(elem)

  case list.find(acc, fn(existing) { to_key(existing) == key }) {
    Ok(existing) -> Error(#(existing, elem))
    Error(Nil) -> Ok([elem, ..acc])
  }
}

/// Like filter_map, but collect all errors, without early returning.
fn traverse_map(
  list: List(a),
  body: fn(a) -> Result(ok, List(err)),
) -> Result(List(ok), List(err)) {
  do_traverse_map(list, body, [], [])
}

fn do_traverse_map(
  list: List(a),
  body: fn(a) -> Result(ok, List(err)),
  oks: List(ok),
  errs: List(err),
) -> Result(List(ok), List(err)) {
  case list {
    [] ->
      case errs {
        [] -> Ok(list.reverse(oks))
        _ -> Error(list.reverse(errs))
      }
    [head, ..tail] ->
      case body(head) {
        Ok(ok) -> do_traverse_map(tail, body, [ok, ..oks], errs)
        Error(err) -> do_traverse_map(tail, body, oks, list.append(err, errs))
      }
  }
}

fn try(
  result: Result(ok, err),
  f: fn(ok) -> Result(ok2, List(err)),
) -> Result(ok2, List(err)) {
  case result {
    Ok(ok) -> f(ok)
    Error(err) -> Error([err])
  }
}
